"""
Ludolph: Monitoring Jabber bot
Copyright (C) 2012-2013 Erigones s. r. o.
This file is part of Ludolph.

See the file LICENSE for copying permission.
"""
import time
import logging

from ludolph.__init__ import __doc__ as ABOUT
from ludolph.__init__ import __version__ as VERSION
from ludolph.command import command, parameter_required, admin_required
from ludolph.plugins.plugin import LudolphPlugin

logger = logging.getLogger(__name__)


class Base(LudolphPlugin):
    """
    Ludolph jabber bot base commands.
    """
    @command
    def help(self, msg, cmdstr=None):
        """
        Show this help.

        Usage: help [command]
        """
        # Global help or command help?
        if cmdstr:
            cmd = self.xmpp.get_command(cmdstr)
            if cmd:
                # Remove whitespaces from __doc__ lines
                desc = '\n'.join(map(str.strip, cmd['doc'].split('\n')))
                # **Command name** (module) + desc
                return '**%s** (%s)\n\n%s' % (cmd['str'], cmd['module'], desc)

        # Create dict with module name as key and list of commands as value
        cmd_map = {}
        for cmd_name in self.xmpp.available_commands():
            cmd = self.xmpp.commands[cmd_name]
            mod_name = cmd['module']

            if not mod_name in cmd_map:
                cmd_map[mod_name] = []

            cmd_map[mod_name].append(cmd_name)

        out = 'List of available Ludolph commands:\n'

        for mod_name, cmd_names in cmd_map.items():
            # Item: module name
            out += '\n* %s\n\n' % mod_name

            for name in cmd_names:
                try:
                    # First line of __doc__
                    desc = self.xmpp.commands[name]['doc'].split('\n')[0]
                    # Lowercase first char and remove trailing dot
                    desc = ' - ' + desc[0].lower() + desc[1:].rstrip('.')
                except IndexError:
                    desc = ''

                # SubItem: line of command + description
                out += '  * **%s**%s\n' % (name, desc)

        out += '\nUse "help <command>" for more information about the command usage'

        return out

    @command
    def version(self, msg):
        """
        Display Ludolph version.

        Usage: version
        """
        return 'Version: ' + VERSION

    @command
    def about(self, msg):
        """
        Details about this project.

        Usage: about
        """
        return ABOUT.strip()

    @admin_required
    @command
    def roster_list(self, msg):
        """
        List of users on Ludolph's roster (admin only).

        Usage: roster-list
        """
        roster = self.xmpp.client_roster
        out = ''

        for i in roster.keys():
            out += '%s\t%s\n' % (i, roster[i]['subscription'])

        return out

    @admin_required
    @parameter_required(1)
    @command
    def roster_remove(self, msg, user):
        """
        Remove user from Ludolph's roster (admin only).

        Usage: roster-remove <JID>
        """
        if user in self.xmpp.client_roster.keys():
            self.xmpp.send_presence(pto=user, ptype='unsubscribe')
            self.xmpp.del_roster_item(user)
            return 'User **' + user + '** removed from roster'
        else:
            return 'User **' + user + '** cannot be removed from roster'

    @command
    def uptime(self, msg):
        """
        Show Ludolph uptime.

        Usage: uptime
        """
        u = time.time() - self.xmpp._start_time
        m, s = divmod(u, 60)
        h, m = divmod(m, 60)
        d, h = divmod(h, 24)

        return 'up %d days, %d hours, %d minutes, %d seconds' % (d, h, m, s)

    @admin_required
    @command
    def muc_invite(self, msg, user=None):
        """
        Invite user or yourself to multi-user chat room (admin only).

        Usage: muc-invite [JID]
        """
        if not self.xmpp.room:
            return 'ERROR: MUC room disabled'

        if not user:
            user = self.xmpp.get_jid(msg)

        self.xmpp.muc.invite(self.xmpp.room, user)

        return 'Inviting **%s** to MUC room %s' % (user, self.xmpp.room)



"""
This is a port of the ruby zabbix api found here:
http://trac.red-tux.net/browser/ruby/api/zbx_api.rb

LGPL 2.1   http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html

Zabbix API Python Library.
Original Ruby Library is Copyright (C) 2009 Andrew Nelson nelsonab(at)red-tux(dot)net
Python Library is Copyright (C) 2009 Brett Lentz brett.lentz(at)gmail(dot)com
                  Copyright (C) 2013 Erigones s. r. o. erigones(at)erigones(dot)com

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA


NOTES:
The API requires zabbix 1.8 or later.
Currently, not all of the API is implemented, and some functionality is broken. This is a work in progress.
"""

import base64
import hashlib
import logging
import string
import re
import datetime
import json
import time

try:
    import urllib2
except ImportError:
    import urllib.request as urllib2  # python3

from collections import deque


PARENT_LOGGER = __name__
DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'
TRIGGER_SEVERITY = (
    'not_classified',
    'information',
    'warning',
    'average',
    'high',
    'disaster',
)
RE_HIDE_AUTH = (
    (re.compile(r'("auth": )".*?"'), r'\1"***"'),
    (re.compile(r'("password": )".*?"'), r'\1"***"'),
)


def hide_auth(msg):
    """Remove sensitive information from msg."""
    for pattern, repl in RE_HIDE_AUTH:
        msg = pattern.sub(repl, msg)

    return msg


class ZabbixAPIException(Exception):
    """
    Generic zabbix api exception.
    code list:
         -32602 - Invalid params (eg already exists)
         -32500 - no permissions
    """
    def __init__(self, message, *args, **kwargs):
        message = hide_auth(message)  # Remove sensitive information
        super(ZabbixAPIException, self).__init__(message, *args, **kwargs)


class AlreadyExists(ZabbixAPIException):
    """
    Zabbix object already exists.
    """
    pass


class ZabbixAPI(object):
    __username__ = None
    __password__ = None
    auth = ''
    params = None
    method = None
    id = 0

    def __init__(self, server='http://localhost/zabbix', user=None, passwd=None,
                 log_level=logging.WARNING, timeout=10, r_query_len=10, **kwargs):
        """
        Create an API object.
        We're going to use proto://server/path to find the JSON-RPC api.
        :param str server: Server to connect to
        :param str path: Path leading to the zabbix install
        :param str user: Optional HTTP auth username
        :param str passwd: Optional HTTP auth password
        :param int log_level: Logging level
        :param int timeout: Timeout for HTTP requests to api
        :param int r_query_len: Max length of query history
        :param **kwargs: Data to pass to each api module
        """
        self.logger = logging.getLogger(PARENT_LOGGER)
        self.set_log_level(log_level)
        self.server = server
        self.url = server + '/api_jsonrpc.php'
        self.proto = server.split('://')[0]
        self.httpuser = user
        self.httppasswd = passwd
        self.timeout = timeout
        self.r_query = deque([], maxlen=r_query_len)

        # sub-class instances
        self.usergroup = ZabbixAPISubClass(self, dict({'prefix': 'usergroup'}, **kwargs))
        self.user = ZabbixAPISubClass(self, dict({'prefix': 'user'}, **kwargs))
        self.host = ZabbixAPISubClass(self, dict({'prefix': 'host'}, **kwargs))
        self.item = ZabbixAPISubClass(self, dict({'prefix': 'item'}, **kwargs))
        self.hostgroup = ZabbixAPISubClass(self, dict({'prefix': 'hostgroup'}, **kwargs))
        self.hostinterface = ZabbixAPISubClass(self, dict({'prefix': 'hostinterface'}, **kwargs))
        self.application = ZabbixAPISubClass(self, dict({'prefix': 'application'}, **kwargs))
        self.trigger = ZabbixAPISubClass(self, dict({'prefix': 'trigger'}, **kwargs))
        self.template = ZabbixAPISubClass(self, dict({'prefix': 'template'}, **kwargs))
        self.action = ZabbixAPISubClass(self, dict({'prefix': 'action'}, **kwargs))
        self.alert = ZabbixAPISubClass(self, dict({'prefix': 'alert'}, **kwargs))
        self.info = ZabbixAPISubClass(self, dict({'prefix': 'info'}, **kwargs))
        self.event = ZabbixAPISubClass(self, dict({'prefix': 'event'}, **kwargs))
        self.graph = ZabbixAPISubClass(self, dict({'prefix': 'graph'}, **kwargs))
        self.graphitem = ZabbixAPISubClass(self, dict({'prefix': 'graphitem'}, **kwargs))
        self.map = ZabbixAPISubClass(self, dict({'prefix': 'map'}, **kwargs))
        self.screen = ZabbixAPISubClass(self, dict({'prefix': 'screen'}, **kwargs))
        self.script = ZabbixAPISubClass(self, dict({'prefix': 'script'}, **kwargs))
        self.usermacro = ZabbixAPISubClass(self, dict({'prefix': 'usermacro'}, **kwargs))
        self.drule = ZabbixAPISubClass(self, dict({'prefix': 'drule'}, **kwargs))
        self.history = ZabbixAPISubClass(self, dict({'prefix': 'history'}, **kwargs))
        self.maintenance = ZabbixAPISubClass(self, dict({'prefix': 'maintenance'}, **kwargs))
        self.proxy = ZabbixAPISubClass(self, dict({'prefix': 'proxy'}, **kwargs))
        self.apiinfo = ZabbixAPISubClass(self, dict({'prefix': 'apiinfo'}, **kwargs))
        self.configuration = ZabbixAPISubClass(self, dict({'prefix': 'configuration'}, **kwargs))
        self.dcheck = ZabbixAPISubClass(self, dict({'prefix': 'dcheck'}, **kwargs))
        self.dhost = ZabbixAPISubClass(self, dict({'prefix': 'dhost'}, **kwargs))
        self.discoveryrule = ZabbixAPISubClass(self, dict({'prefix': 'discoveryrule'}, **kwargs))
        self.dservice = ZabbixAPISubClass(self, dict({'prefix': 'dservice'}, **kwargs))
        self.iconmap = ZabbixAPISubClass(self, dict({'prefix': 'iconmap'}, **kwargs))
        self.image = ZabbixAPISubClass(self, dict({'prefix': 'image'}, **kwargs))
        self.mediatype = ZabbixAPISubClass(self, dict({'prefix': 'mediatype'}, **kwargs))
        self.service = ZabbixAPISubClass(self, dict({'prefix': 'service'}, **kwargs))
        self.templatescreen = ZabbixAPISubClass(self, dict({'prefix': 'templatescreen'}, **kwargs))
        self.usermedia = ZabbixAPISubClass(self, dict({'prefix': 'usermedia'}, **kwargs))
        self.hostinterface = ZabbixAPISubClass(self, dict({'prefix': 'hostinterface'}, **kwargs))
        self.triggerprototype = ZabbixAPISubClass(self, dict({'prefix': 'triggerprototype'}, **kwargs))
        self.graphprototype = ZabbixAPISubClass(self, dict({'prefix': 'graphprototype'}, **kwargs))
        self.itemprototype = ZabbixAPISubClass(self, dict({'prefix': 'itemprototype'}, **kwargs))
        self.webcheck = ZabbixAPISubClass(self, dict({'prefix': 'webcheck'}, **kwargs))

        self.debug('url: %s', self.url)

    def set_log_level(self, level):
        self.debug('Set logging level to %d', level)
        self.logger.setLevel(level)

    @classmethod
    def get_severity(cls, prio):
        """
        Return severity string from severity id.
        """
        try:
            return TRIGGER_SEVERITY[int(prio)]
        except IndexError:
            return 'unknown'

    @classmethod
    def get_datetime(cls, timestamp):
        """
        Return python datetime object from unix timestamp.
        """
        return datetime.datetime.fromtimestamp(int(timestamp))

    @classmethod
    def convert_datetime(cls, dt, dt_format=DATETIME_FORMAT):
        """
        Convert python datetime to human readable date and time string.
        """
        return dt.strftime(dt_format)

    @classmethod
    def timestamp_to_datetime(cls, dt, dt_format=DATETIME_FORMAT):
        """
        Convert unix timestamp to human readable date/time string.
        """
        return cls.convert_datetime(cls.get_datetime(dt), dt_format=dt_format)

    @classmethod
    def get_age(cls, dt):
        """
        Calculate delta between current time and datetime and return a human readable form of the delta object.
        """
        delta = datetime.datetime.now() - dt
        days = delta.days
        hours, rem = divmod(delta.seconds, 3600)
        minutes, seconds = divmod(rem, 60)

        if days:
            return '%dd %dh %dm' % (days, hours, minutes)
        else:
            return '%dh %dm %ds' % (hours, minutes, seconds)

    def recent_query(self):
        """
        Return recent query.
        """
        return list(self.r_query)

    def log(self, level, msg, *args):
        return self.logger.log(level, msg, *args)

    def debug(self, msg, *args):
        return self.log(logging.DEBUG, msg, *args)

    def json_obj(self, method, params=None):
        if params is None:
            params = {}

        obj = {
            'jsonrpc': '2.0',
            'method': method,
            'params': params,
            'auth': self.auth,
            'id': self.id
        }
        self.debug('json_obj: %s', obj)

        return json.dumps(obj)

    def login(self, user=None, password=None, save=True):
        if user and password:
            l_user = user
            l_password = password
            if save:
                self.__username__ = user
                self.__password__ = password

        elif self.__username__ and self.__password__:
            l_user = self.__username__
            l_password = self.__password__

        else:
            raise ZabbixAPIException('No authentication information available.')

        # Don't print the raw password.
        hashed_pw_string = 'md5(' + hashlib.md5(l_password.encode('utf-8')).hexdigest() + ')'
        self.debug('Trying to login with %s:%s', repr(l_user), repr(hashed_pw_string))
        obj = self.json_obj('user.authenticate', {'user': l_user, 'password': l_password})
        result = self.do_request(obj)
        self.auth = result['result']

    def logged_in(self):
        return bool(self.auth)

    def test_login(self):
        if self.auth:
            obj = self.json_obj('user.checkAuthentication', {'sessionid': self.auth})
            result = self.do_request(obj)

            if 'result' in result and result['result']:
                return True  # auth hash good
            else:
                self.auth = ''  # auth hash bad

        return False

    def do_request(self, json_obj):
        headers = {
            'Content-Type': 'application/json-rpc',
            'User-Agent': 'python/zabbix_api',
        }

        if self.httpuser:
            self.debug('HTTP Auth enabled')
            auth = 'Basic ' + string.strip(base64.encodestring(self.httpuser + ':' + self.httppasswd))
            headers['Authorization'] = auth

        self.r_query.append(str(json_obj))
        self.debug('Sending: %s', json_obj)
        self.debug('Sending headers: %s', headers)

        request = urllib2.Request(url=self.url, data=json_obj.encode('utf-8'), headers=headers)

        if self.proto == 'https':
            http_handler = urllib2.HTTPSHandler(debuglevel=0)
        elif self.proto == 'http':
            http_handler = urllib2.HTTPHandler(debuglevel=0)
        else:
            raise ZabbixAPIException('Unknown protocol %s' % self.proto)

        opener = urllib2.build_opener(http_handler)
        urllib2.install_opener(opener)

        try:
            response = opener.open(request, timeout=self.timeout)
        except Exception as e:
            raise ZabbixAPIException('HTTP connection problem: ' + str(e))

        self.debug('Response Code: %s', response.code)

        # NOTE: Getting a 412 response code means the headers are not in the list of allowed headers.
        if response.code != 200:
            raise ZabbixAPIException('HTTP error %s: %s' % (response.status, response.reason))

        reads = response.read()

        if len(reads) == 0:
            raise ZabbixAPIException('Received zero answer')

        try:
            jobj = json.loads(reads.decode('utf-8'))
        except ValueError as e:
            self.log(logging.ERROR, 'Unable to decode. returned string: %s', reads)
            raise ZabbixAPIException('Unable to decode response: ' + str(e))

        self.debug('Response Body: %s', jobj)
        self.id += 1

        if 'error' in jobj:  # some exception
            msg = 'Error %s: %s, %s while sending %s' % (jobj['error']['code'], jobj['error']['message'],
                                                         jobj['error']['data'], str(json_obj))

            if re.search('.*already\sexists.*', jobj['error']['data'], re.I):  # already exists
                raise AlreadyExists(msg, jobj['error']['code'])
            else:
                raise ZabbixAPIException(msg, jobj['error']['code'])

        return jobj

    def api_version(self, **options):
        self.check_auth()
        obj = self.do_request(self.json_obj('APIInfo.version', options))

        return obj['result']

    def check_auth(self):
        if not self.logged_in():
            raise ZabbixAPIException('Not logged in.')


class ZabbixAPISubClass(object):
    """
    Wrapper class to ensure all calls go through the parent object.
    """
    def __init__(self, parent, data, **kwargs):
        self.data = data
        self.parent = parent
        self.parent.debug('Creating %s', self.__class__.__name__)

        # Save any extra info passed in
        for key, val in kwargs.items():
            setattr(self, key, val)

    def __getattr__(self, name):
        if self.data['prefix'] == 'configuration' and name == 'import_':  # workaround for "import" method
            name = 'import'

        def method(*opts):
            return self.universal('%s.%s' % (self.data['prefix'], name), opts[0])
        return method

    def universal(self, method, opts):
        """
        Check authentication and perform actual API request and re-login if needed.
        """
        start_time = time.time()
        self.parent.check_auth()
        self.parent.log(logging.INFO, '[%s-%05d] Calling Zabbix API method "%s"', start_time, self.parent.id, method)
        self.parent.log(logging.DEBUG, '\twith options: %s', opts)

        try:
            return self.parent.do_request(self.parent.json_obj(method, opts))['result']
        except ZabbixAPIException as ex:
            if str(ex).find('Not authorized while sending') >= 0:
                self.parent.log(logging.WARNING, 'Zabbix API not logged in (%s). Performing Zabbix API re-login', ex)
                try:
                    self.parent.auth = ''  # reset auth before re-login
                    self.parent.login()
                except ZabbixAPIException as e:
                    self.parent.log(logging.ERROR, 'Zabbix API login error (%s)', e)
                    self.parent.auth = ''  # logged_in() will always return False
                    raise e
                else:
                    return self.parent.do_request(self.parent.json_obj(method, opts))['result']
            else:
                raise ex
        finally:
            self.parent.log(logging.INFO, '[%s-%05d] Zabbix API method "%s" finished in %g seconds',
                            start_time, self.parent.id, method, (time.time() - start_time))



"""
Ludolph: Monitoring Jabber bot
Copyright (C) 2012-2013 Erigones s. r. o.
This file is part of Ludolph.

See the file LICENSE for copying permission.
"""
import logging
from datetime import datetime, timedelta

from ludolph.command import command, parameter_required
from ludolph.message import red, green
from ludolph.plugins.plugin import LudolphPlugin
from ludolph.plugins.zabbix_api import ZabbixAPI, ZabbixAPIException

logger = logging.getLogger(__name__)

TIMEOUT = 10
DUTY_GROUP = 'On-Call Duty'


def zabbix_command(f):
    """
    Decorator for executing zabbix API commands and checking zabbix API errors.
    """
    def wrap(obj, msg, *args, **kwargs):
        def api_error(errmsg='Zabbix API not available'):
            # Log and reply with error message
            logger.error(errmsg)
            obj.xmpp.msg_reply(msg, 'ERROR: ' + errmsg)
            return None

        # Was never logged in. Repair authentication and restart Ludolph.
        if not obj.zapi.logged_in():
            return api_error()

        try:
            return f(obj, msg, *args, **kwargs)
        except ZabbixAPIException as ex:
            # API command problem
            return api_error('Zabbix API error (%s)' % ex)

    return wrap


class Zabbix(LudolphPlugin):
    """
    Zabbix API connector for LudolphBot.

    Zabbix >= 2.0.6 is required.
    https://www.zabbix.com/documentation/2.0/manual/appendix/api/api
    """
    zapi = None

    def __init__(self, config, **kwargs):
        """
        Login to zabbix.
        """
        self.init(config)

    def init(self, config):
        """
        Initialize zapi and try to login.
        """
        # HTTP authentication?
        httpuser = None
        httppasswd = None
        if config.has_option('zabbix', 'httpuser'):
            httpuser = config.get('zabbix', 'httpuser')
        if config.has_option('zabbix', 'httppasswd'):
            httppasswd = config.get('zabbix', 'httppasswd')

        self.zapi = ZabbixAPI(server=config.get('zabbix', 'server'),
                              user=httpuser, passwd=httppasswd, timeout=TIMEOUT,
                              log_level=logging.getLevelName(config.get('global', 'loglevel')))

        # Login and save zabbix credentials
        try:
            logger.info('Zabbix API login')
            self.zapi.login(config.get('zabbix', 'username'),
                            config.get('zabbix', 'password'), save=True)
        except ZabbixAPIException as e:
            logger.critical('Zabbix API login error (%s)', e)

    @zabbix_command
    @command
    def zabbix_version(self, msg):
        """
        Show version of Zabbix API.

        Usage: zabbix-version
        """
        return 'Zabbix API version: ' + self.zapi.api_version()

    @zabbix_command
    @command
    def alerts(self, msg, notes=False):
        """
        Show a list of current zabbix alerts.

        Use optional "notes" parameter to display all notes attached to each vent ID.

        Usage: alerts [notes]
        """
        out = ''
        # Get triggers
        triggers = self.zapi.trigger.get({
                'groupids': None,
                'hostids': None,
                'monitored': True,
                'maintenance': False,
                'skipDependent': True,
                'filter': {'priority': None, 'value': 1},  # TRIGGER_VALUE_TRUE
                'selectHosts': ['hostid', 'name', 'maintenance_status', 'maintenance_type', 'maintenanceid'],
                'selectLastEvent':  'extend',  # API_OUTPUT_EXTEND
                'output': ['triggerid', 'value_flags', 'error', 'url', 'expression', 'description', 'priority', 'type', 'comments'],
                'sortfield': 'lastchange',
                'sortorder': 'DESC',  # ZBX_SORT_DOWN
        })

        # Get notes = event acknowledges
        if notes:
            events = self.zapi.event.get({
                    'eventids': [t['lastEvent']['eventid'] for t in triggers if t['lastEvent']],
                    'output': 'extend',
                    'select_acknowledges': 'extend',
                    'sortfield': 'eventid',
                    'sortorder': 'DESC',
            })

        for trigger in triggers:
            # If trigger is lost (broken expression) we skip it
            if not trigger['hosts']:
                continue

            # Event
            event = trigger['lastEvent']
            if event:
                eventid = event['eventid']
                # Ack
                if int(event['acknowledged']):
                    ack = '^^**ACK**^^'
                else:
                    ack = ''
            else:
                # WTF?
                eventid = '????'
                ack = ''

            # Host and hostname
            host = trigger['hosts'][0]
            hostname = host['name']
            if int(host['maintenance_status']):
                hostname += ' **++**'  # some kind of maintenance

            # Trigger description
            desc = str(trigger['description'])
            if trigger['error'] or int(trigger['value_flags']):
                desc += ' **??**'  # some kind of trigger error

            # Priority
            prio = self.zapi.get_severity(trigger['priority']).ljust(12)

            # Last change and age
            dt = self.zapi.get_datetime(trigger['lastchange'])
            #last = self.zapi.convert_datetime(dt)
            age = '^^%s^^' % self.zapi.get_age(dt)

            comments = ''
            if trigger['error']:
                comments += '\n\t\t^^**Error:** %s^^' % trigger['error']

            if trigger['comments']:
                comments += '\n\t\t^^%s^^' % trigger['comments'].strip()

            acknowledges = ''
            if notes:
                for i, e in enumerate(events):
                    if e['eventid'] == event['eventid']:
                        for a in e['acknowledges']:
                            acknowledges = '\n\t\t__%s: %s__' % (self.zapi.get_datetime(a['clock']), a['message']) + acknowledges
                        del events[i]
                        break

            out += '**%s**\t%s\t%s\t%s\t%s\t%s%s%s\n' % (eventid, prio, hostname, desc, age, ack, comments, acknowledges)

        out += '\n**%d** issues are shown.\n%s/tr_status.php?groupid=0&hostid=0' % (len(triggers), self.zapi.server)

        return out

    @zabbix_command
    @parameter_required(1)
    @command
    def ack(self, msg, eventid, *note):
        """
        Acknowledge event with optional note.

        Usage: ack <event ID> [note]
        """
        try:
            eventid = int(eventid)
        except ValueError:
            return 'ERROR: Integer required'

        message = '%s: ' % self.xmpp.get_jid(msg)

        if note:
            message += ' '.join(note)
        else:
            message += 'ack'

        self.zapi.event.acknowledge({
            'eventids': [eventid],
            'message': message,
        })

        return 'Event ID **%s** acknowledged' % eventid

    def _outage_del(self, msg, mid):
        """
        Delete maintenance period specified by maintenance ID.

        Usage: outage-del <maintenance ID>
        """
        try:
            mid = int(mid)
        except ValueError:
            return 'ERROR: Integer required'

        self.zapi.maintenance.delete([mid])

        return 'Maintenance ID **%s** deleted' % mid

    def _outage_add(self, msg, host_or_group, duration):
        """
        Set maintenance period for specified host and time.

        Usage: outage-add <host/group name> <duration in minutes>
        """
        # Get start and end time
        try:
            duration = int(duration)
        except ValueError:
            return 'ERROR: Integer required'

        period = timedelta(minutes=duration)
        _now = datetime.now()
        _end = _now + period
        now = _now.strftime('%s')
        end = _end.strftime('%s')
        jid = self.xmpp.get_jid(msg)
        options = {
                'active_since': now,
                'active_till': end,
                'description': str(jid),
                'maintenance_type': 0,  # with data collection
                'timeperiods': [{
                    'timeperiod_type': 0,  # one time only
                    'start_date': now,
                    'period': period.seconds,
                }],
        }

        # Get hosts
        hosts = self.zapi.host.get({
            'search': {'name': host_or_group},
            'output': ['hostid', 'name'],
        })

        if hosts:
            options['hostids'] = [i['hostid'] for i in hosts]
            names = [i['name'] for i in hosts]

        if not hosts:
            # Get groups
            groups = self.zapi.hostgroup.get({
                'search': {'name': host_or_group},
                'output': ['groupids', 'name'],
            })

            if groups:
                options['groupids'] = [i['groupid'] for i in groups]
                names = [i['name'] for i in groups]
            else:
                return "ERROR: Host/Group not found"

        names = ', '.join(names)
        options['name'] = 'Maintenance for %s - %s' % (names, now)

        # Create maintenance period
        res = self.zapi.maintenance.create(options)

        return 'Added maintenance ID **%s** for %s %s' % (res['maintenanceids'][0],
                                                        'host' if hosts else 'group', names)

    @zabbix_command
    @command
    def outage(self, msg, mid_or_host_or_group=None, duration=None):
        """
        Show, create or delete maintenance periods.

        Show all maintenance periods.
        Usage: outage

        Set maintenance period for specified host and time.
        Usage: outage <host/group name> <duration in minutes>

        Delete maintenance period specified by maintenance ID.
        Usage: outage <maintenance ID>
        """
        if duration:
            return self._outage_add(msg, mid_or_host_or_group, duration)
        elif mid_or_host_or_group:
            return self._outage_del(msg, mid_or_host_or_group)

        out = ''
        # Display list of maintenances
        maintenances = self.zapi.maintenance.get({
            'output': 'extend',
            'sortfield': ['maintenanceid', 'name'],
            'sortorder': 'ASC',
            'selectHosts': 'extend',
            'selectGroups': 'extend',
        })

        for i in maintenances:
            if i['hosts']:
                hosts = '\n\t^^hosts:\t%s^^' % ', '.join([h['name'] for h in i['hosts']])
            else:
                hosts = ''

            if i['groups']:
                groups = '\n\t^^groups:\t%s^^' % ', '.join([g['name'] for g in i['groups']])
            else:
                groups = ''

            since = self.zapi.timestamp_to_datetime(i['active_since'])
            until = self.zapi.timestamp_to_datetime(i['active_till'])
            out += '**%s**\t%s\t%s%s%s\n\t%s - %s\n' % (i['maintenanceid'], i['name'], i['description'], hosts, groups, since, until)

        out += '\n**%d** maintenances are shown.\n%s' % (len(maintenances), self.zapi.server + '/maintenance.php?groupid=0')

        return out

    @zabbix_command
    @command
    def hosts(self, msg, hoststr=None):
        """
        Show a list of hosts.

        Usage: hosts [host name search string]
        """
        out = ''
        params = {
            'output': ['hostid', 'name', 'available', 'maintenance_status', 'status'],
            'selectInventory': 1,  # All inventory items
            'sortfield': ['name', 'hostid'],
            'sortorder': 'ASC',
        }

        if hoststr:
            params['search'] = {'name': hoststr}

        # Get hosts
        hosts = self.zapi.host.get(params)

        for host in hosts:
            if int(host['maintenance_status']):
                host['name'] += ' **++**'  # some kind of maintenance

            if int(host['status']):
                status = 'Not monitored'
            else:
                status = 'Monitored'

            ae = int(host['available'])
            available = 'Z'
            if ae == 1:
                available = green('Z')
            elif ae == 2:
                available = red('Z')

            _inventory = []
            if host['inventory']:
                for key, val in host['inventory'].items():
                    if val and key not in ('inventory_mode', 'hostid'):
                        _inventory.append('**%s**: %s' % (key, val))

            if _inventory:
                inventory = '\n\t\t^^%s^^' % str(', '.join(_inventory)).strip()
            else:
                inventory = ''

            out += '**%s**\t%s\t%s\t%s%s\n' % (host['hostid'], host['name'], status, available, inventory)

        out += '\n**%d** hosts are shown.\n%s/hosts.php?groupid=0' % (len(hosts), self.zapi.server)

        return out

    @zabbix_command
    @command
    def groups(self, msg, groupstr=None):
        """
        Show a list of host groups.

        Usage: groups [group name search string]
        """
        out = ''
        params = {
            'output': ['groupid', 'name'],
            'selectHosts': ['hostid', 'name'],
            'sortfield': ['name', 'groupid'],
            'sortorder': 'ASC',
        }

        if groupstr:
            params['search'] = {'name': groupstr}

        # Get groups
        groups = self.zapi.hostgroup.get(params)

        for group in groups:
            _hosts = ['**%s**: %s' % (h['hostid'], h['name']) for h in group['hosts'] if h]
            hosts = '\n\t\t^^%s ^^' % ', '.join(_hosts)
            out += '**%s**\t%s%s\n' % (group['groupid'], group['name'], hosts)

        out += '\n**%d** hostgroups are shown.\n%s/hostgroups.php' % (len(groups), self.zapi.server)

        return out

    @zabbix_command
    @command
    def duty(self, msg):
        """
        Show a list of users in duty user group.

        Usage: duty
        """
        out = ''
        # Get group
        duty = self.zapi.usergroup.get({
                'filter': {'name': DUTY_GROUP},
                'output': ['usrgrpid', 'name', 'users_status'],
                'selectUsers': 'extend',  # API_OUTPUT_EXTEND
        })

        if not len(duty):
            return 'ERROR: Duty user group ("%s") not found' % DUTY_GROUP

        for u in duty[0]['users']:
            if int(u['users_status']):
                status = red('disabled')
            else:
                status = green('enabled')

            out += '%s\t%s\n' % (u['alias'], status)

        out += '\n**%d** users in duty group.' % len(duty[0]['users'])

        return out



"""
Ludolph: Monitoring Jabber bot
Copyright (C) 2012-2013 Erigones s. r. o.
This file is part of Ludolph.

See the file LICENSE for copying permission.
"""


class LudolphPlugin(object):
    """
    Ludolph plugin base class.
    """
    xmpp = None  # Reference to LudolphBot object

    def __init__(self, config, reinit=False, **kwargs):
        pass  # Implement plugin initialization here



"""
Ludolph: Monitoring Jabber Bot
Version: 0.4.0
Homepage: https://github.com/erigones/Ludolph
Copyright (C) 2012-2013 Erigones s. r. o.

See the LICENSE file for copying permission.
"""

__version__ = '0.4.0'



"""
Ludolph: Monitoring Jabber Bot
Copyright (C) 2012-2013 Erigones s. r. o.
This file is part of Ludolph.

See the LICENSE file for copying permission.
"""
from logging import getLogger

logger = getLogger(__name__)

COMMANDS = {}  # command : {name, module, doc}
USERS = set()  # List of users
ADMINS = set() # List of admins

__all__ = ['command', 'parameter_required', 'admin_required']


def command(f):
    """
    Decorator for registering available commands.
    """
    global COMMANDS
    global USERS

    def wrap(obj, msg, *args, **kwargs):
        user = obj.xmpp.get_jid(msg)
        cmd = '%s.%s' % (f.__module__, f.__name__)

        if not USERS or user in USERS:
            logger.info('User "%s" requested command "%s" (%s)', user, msg['body'], cmd)
            # Parse optional parameters
            if f.func_defaults:
                required_pos = len(args) + 1
                optional_end = required_pos + len(f.func_defaults)
                optional_args = msg['body'].strip().split()[required_pos:optional_end]
                args += tuple(optional_args)

            # Reply with function output
            out = f(obj, msg, *args, **kwargs)
            logger.debug('Command output: "%s"', out)
            obj.xmpp.msg_reply(msg, out)
            return True
        else:
            logger.warning('Unauthorized command "%s" (%s) from "%s"', msg['body'], cmd, user)
            obj.xmpp.msg_reply(msg, 'ERROR: Permission denied')
            return None

    # Create command name - skip methods which start with underscore
    if f.__name__.startswith('_'):
        # Not a public command, but we will execute the method (private helper)
        return wrap
    else:
        name = f.__name__.replace('_', '-')

    # Check if command exists
    if name in COMMANDS.keys():
        logger.critical('Command "%s" from plugin "%s" overlaps with existing command from module "%s"',
                        name, f.__module__, COMMANDS[name]['module'])
        return None

    # Save module and method name
    COMMANDS[name] = {'name': f.__name__, 'module': f.__module__}

    # Save documentation
    if f.__doc__:
        COMMANDS[name]['doc'] = f.__doc__.strip()
    else:
        logger.error('Missing documentation for command "%s"', name)
        COMMANDS[name]['doc'] = ''

    return wrap


def parameter_required(count=1):
    """
    Decorator for required command parameters.
    """
    def parameter_required_decorator(f):
        def wrap(obj, msg, *args, **kwargs):
            #Try to get command parameter
            params = msg['body'].strip().split()[1:]

            if len(params) < count:
                user = obj.xmpp.get_jid(msg)
                logger.warning('Missing parameter in command "%s" from user "%s"', msg['body'], user)
                obj.xmpp.msg_reply(msg, 'ERROR: Missing parameter')
                return None
            else:
                params.extend(args)
                return f(obj, msg, *params, **kwargs)

        return wrap
    return parameter_required_decorator


def admin_required(f):
    """
    Decorator for admin only commands.
    """
    global ADMINS

    def wrap(obj, msg, *args, **kwargs):
        user = obj.xmpp.get_jid(msg)

        if not ADMINS or user in ADMINS:
            return f(obj, msg, *args, **kwargs)
        else:
            logger.warning('Unauthorized command "%s" from user "%s"', msg['body'], user)
            obj.xmpp.msg_reply(msg, 'ERROR: Permission denied')
            return None

    return wrap



"""
Ludolph: Monitoring Jabber Bot
Copyright (C) 2012-2013 Erigones s. r. o.
This file is part of Ludolph.

See the LICENSE file for copying permission.
"""

import os
import sys
import signal
import logging

# In order to make sure that Unicode is handled properly
# in Python 2.x, reset the default encoding.
if sys.version_info[0] < 3:
    from ConfigParser import RawConfigParser
else:
    from configparser import RawConfigParser

from ludolph.bot import LudolphBot
from ludolph.__init__ import __version__ as VERSION

LOGFORMAT = '%(asctime)s %(levelname)-8s %(name)s: %(message)s'

logger = logging.getLogger(__name__)


def daemonize():
    """
    http://code.activestate.com/recipes/278731-creating-a-daemon-the-python-way/
    http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/
    """
    try:
        pid = os.fork()  # Fork #1
        if pid > 0:
            sys.exit(0)  # Exit first parent
    except OSError as e:
        sys.stderr.write('Fork #1 failed: %d (%s)\n' % (e.errno, e.strerror))
        sys.exit(1)

    # The first child. Decouple from parent environment
    # Become session leader of this new session.
    # Also be guaranteed not to have a controlling terminal
    os.chdir('/')
    os.setsid()
    os.umask(0)

    try:
        pid = os.fork()  # Fork #2
        if pid > 0:
            sys.exit(0)  # Exit from second parent
    except OSError as e:
        sys.stderr.write('Fork #2 failed: %d (%s)\n' % (e.errno, e.strerror))
        sys.exit(1)

    # Close all open file descriptors
    import resource  # Resource usage information
    maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
    if maxfd == resource.RLIM_INFINITY:
        maxfd = 1024

    # Iterate through and close all file descriptors
    for fd in range(0, maxfd):
        try:
            os.close(fd)
        except OSError:  # ERROR, fd wasn't open (ignored)
            pass

    # Redirect standard file descriptors to /dev/null
    sys.stdout.flush()
    sys.stderr.flush()
    si = open(os.devnull, 'r')
    so = open(os.devnull, 'a+')
    se = open(os.devnull, 'a+')
    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())

    return 0


def start():
    """
    Start the daemon.
    """
    ret = 0
    cfg = 'ludolph.cfg'
    cfg_fp = None
    cfg_lo = ((os.path.expanduser('~'), '.' + cfg), (sys.prefix, 'etc', cfg), ('/etc', cfg))
    config_base_sections = ('global', 'xmpp')

    # Try to read config file from ~/.ludolph.cfg or /etc/ludolph.cfg
    for i in cfg_lo:
            try:
                cfg_fp = open(os.path.join(*i))
            except IOError:
                continue
            else:
                break

    if not cfg_fp:
        sys.stderr.write("""\nLudolph can't start!\n
You need to create a config file in one this locations: \n%s\n
You can rename ludolph.cfg.example and update the required variables.
The example file is located in: %s\n\n""" % (
            '\n'.join([os.path.join(*i) for i in cfg_lo]),
            os.path.dirname(os.path.abspath(__file__))))
        sys.exit(1)

    # Read and parse configuration
    def load_config(fp, reopen=False):
        config = RawConfigParser()
        if reopen:
            fp = open(fp.name)
        config.readfp(fp)
        return config
    config = load_config(cfg_fp)

    # Prepare logging configuration
    logconfig = {
        'level': logging.getLevelName(config.get('global', 'loglevel')),
        'format': LOGFORMAT,
    }

    if config.has_option('global', 'logfile'):
        logfile = config.get('global', 'logfile').strip()
        if logfile:
            logconfig['filename'] = logfile

    # Daemonize
    if config.has_option('global', 'daemon'):
        if config.getboolean('global', 'daemon'):
            ret = daemonize()
            # Save pid file
            try:
                with open(config.get('global', 'pidfile'), 'w') as fp:
                    fp.write('%s' % os.getpid())
            except Exception as ex:
                # Setup logging just to show this error
                logging.basicConfig(**logconfig)
                logger.critical('Could not write to pidfile (%s)\n', ex)
                sys.exit(1)

    # Setup logging
    logging.basicConfig(**logconfig)

    # All exceptions will be logged
    def log_except_hook(*exc_info):
        logger.critical('Unhandled exception!', exc_info=exc_info)
        #sys.exit(99)
    sys.excepthook = log_except_hook

    # Default configuration
    pipe_file = config.get('global', 'pipe_file')
    pipe_mode = '0600'
    use_tls = True
    use_ssl = False
    address = []

    # Starting
    logger.info('Starting Ludolph %s', VERSION)
    logger.info('Loaded configuration from %s', cfg_fp.name)

    # Load plugins
    def load_plugins(config, reinit=False):
        plugins = {}

        if reinit:
            logger.info('Reinitializing commands')
            from ludolph.command import COMMANDS
            COMMANDS.clear()

        for plugin in config.sections():
            plugin = plugin.lower().strip()
            if plugin in config_base_sections:
                continue
            logger.info('Loading plugin: %s', plugin)
            try:
                clsname = plugin[0].upper() + plugin[1:]
                modname = 'ludolph.plugins.' + plugin
                module = __import__(modname, fromlist=[clsname])
                if reinit:
                    reload(module)
                plugins[modname] = getattr(module, clsname)
            except Exception as ex:
                logger.critical('Could not load plugin %s', plugin)
                logger.exception(ex)

        return plugins
    plugins = load_plugins(config)

    # XMPP connection settings
    if config.has_option('xmpp', 'host'):
        address = [config.get('xmpp', 'host'), '5222']
        if config.has_option('xmpp', 'port'):
            address[1] = config.get('xmpp', 'port')
        logger.info('Connecting to jabber server %s', ':'.join(address))
    else:
        logger.info('Using DNS SRV lookup to find jabber server')

    if config.has_option('xmpp', 'tls'):
        use_tls = config.getboolean('xmpp', 'tls')

    if config.has_option('xmpp', 'ssl'):
        use_ssl = config.getboolean('xmpp', 'ssl')

    # Create pipe file with desired permissions
    if config.has_option('global', 'pipe_mode'):
        pipe_mode = config.get('global', 'pipe_mode')

    logger.info('Creating pipe file %s', pipe_file)
    try:
        os.remove(pipe_file)
    except os.error:
        pass

    os.mkfifo(pipe_file, int(pipe_mode, 8))

    # Here we go
    try:
        xmpp = LudolphBot(config, plugins=plugins)

        def sighup(signalnum, handler):
            config = load_config(cfg_fp, reopen=True)
            logger.info('Reloaded configuration from %s', cfg_fp.name)
            plugins = load_plugins(config, reinit=True)
            xmpp.reload(config, plugins=plugins)

        signal.signal(signal.SIGINT, xmpp.shutdown)
        signal.signal(signal.SIGTERM, xmpp.shutdown)
        signal.signal(signal.SIGHUP, sighup)
        #signal.siginterrupt(signal.SIGHUP, false)  # http://stackoverflow.com/a/4302037

        if xmpp.connect(tuple(address), use_tls=use_tls, use_ssl=use_ssl):
            xmpp.process(block=True)
            sys.exit(ret)
        else:
            logger.error('Ludolph is unable to connect to jabber server')
            sys.exit(2)

    finally:
        # Cleanup
        logger.info('Removing pipe file %s', pipe_file)
        os.remove(pipe_file)


if __name__ == '__main__':
    start()



"""
Ludolph: Monitoring Jabber Bot
Copyright (C) 2012-2013 Erigones s. r. o.
This file is part of Ludolph.

See the LICENSE file for copying permission.
"""

import os
import ssl
import time
import logging
from sleekxmpp import ClientXMPP
from sleekxmpp.xmlstream import ET

from ludolph.message import LudolphMessage
from ludolph.command import COMMANDS, USERS, ADMINS

logger = logging.getLogger(__name__)

__all__ = ['LudolphBot']


class LudolphBot(ClientXMPP):
    """
    Ludolph bot.
    """
    _start_time = None
    _commands = None  # Cached sorted list of commands
    _muc_ready = False
    commands = COMMANDS
    plugins = None
    users = USERS
    admins = ADMINS
    room = None
    room_config = None
    room_users = set()
    room_admins = set()
    muc = None
    nick = 'Ludolph'
    pipe_file = None
    xmpp = None
    maxhistory = '1024'

    def __init__(self, config, plugins=None, *args, **kwargs):
        self._load_config(config)
        logger.info('Initializing *%s* jabber bot', self.nick)
        self._load_plugins(config, plugins)

        # Initialize the SleekXMPP client
        ClientXMPP.__init__(self, config.get('xmpp', 'username'), config.get('xmpp', 'password'))

        # Auto-authorize is enabled by default. User subscriptions are
        # controlled by self._handle_new_subscription
        self.auto_authorize = True

        # Register XMPP plugins
        self.register_plugin('xep_0030')  # Service Discovery
        self.register_plugin('xep_0045')  # Multi-User Chat
        self.register_plugin('xep_0199')  # XMPP Ping

        # Register event handlers
        self.add_event_handler('session_start', self.session_start, threaded=True)
        self.add_event_handler('message', self.message, threaded=True)

        if self.room:
            self.muc = self.plugin['xep_0045']
            self.add_event_handler('groupchat_message', self.muc_message, threaded=True)
            self.add_event_handler('muc::%s::got_online' % self.room, self.muc_online, threaded=True)

        # Start the monitoring thread for reading the pipe file
        self._start_thread('mon_thread', self.mon_thread)

        # Save start time
        self._start_time = time.time()

    def _load_config(self, config):
        """
        Load bot settings from config object.
        """
        logger.info('Configuring jabber bot')

        # Global stuff
        self.pipe_file = config.get('global', 'pipe_file')

        # Get nick name
        if config.has_option('xmpp', 'nick'):
            nick = config.get('xmpp', 'nick').strip()
            if nick:
                self.nick = nick

        # If you are working with an OpenFire server, you will
        # need to use a different SSL version:
        if config.has_option('xmpp', 'sslv3') and config.getboolean('xmpp', 'sslv3'):
            self.ssl_version = ssl.PROTOCOL_SSLv3

        # Read comma-separated option and return a list of JIDs
        def read_jid_array(section, option, keywords=()):
            jids = set()

            if config.has_option(section, option):
                for jid in config.get(section, option).strip().split(','):
                    jid = jid.strip()

                    if not jid:
                        continue

                    if '@' in jid:
                        if jid.startswith('@'):
                            kwd = jid[1:]
                            if kwd in keywords:
                                jids.update(getattr(self, kwd))
                            else:
                                logger.warn('Skipping invalid keyword "%s" from setting "%s"', jid, option)
                        else:
                            jids.add(jid)
                    else:
                        logger.warn('Skipping invalid JID "%s" from setting "%s"', jid, option)

            return jids

        # Users
        self.users.clear()
        self.users.update(read_jid_array('xmpp', 'users'))
        logger.info('Current users: %s', ', '.join(self.users))

        # Admins
        self.admins.clear()
        self.admins.update(read_jid_array('xmpp', 'admins', keywords=('users',)))
        logger.info('Current admins: %s', ', '.join(self.admins))

        # Admins vs. users
        if not self.admins.issubset(self.users):
            for i in self.admins.difference(self.users):
                logger.error('Admin "%s" is not specified in users. '
                             'This may lead to unexpected behaviour. ', i)

        # MUC room
        if config.has_option('xmpp', 'room'):
            self.room = config.get('xmpp', 'room').strip()

        # MUC room users
        self.room_users.clear()
        if self.room:
            self.room_users.update(read_jid_array('xmpp', 'room_users', keywords=('users', 'admins')))
            logger.info('Current room users: %s', ', '.join(self.room_users))

        # MUC room admins
        self.room_admins.clear()
        if self.room:
            self.room_admins.update(read_jid_array('xmpp', 'room_admins', keywords=('users', 'admins', 'room_users')))
            logger.info('Current room admins: %s', ', '.join(self.room_admins))

        # Room admins vs. users
        if not self.room_admins.issubset(self.room_users):
            for i in self.room_admins.difference(self.room_users):
                logger.error('Room admin "%s" is not specified in room_users. '
                             'This may lead to unexpected behaviour. ', i)

    def _load_plugins(self, config, plugins, init=True):
        """
        Initialize plugins.
        The init parameter indicates whether this is a first-time initialization or a reload.
        """
        if init:
            # First-time plugin initialization -> include ourself to plugins dict
            self.xmpp = self
            self.plugins = {__name__: self}
        else:
            # Bot reload - remove disabled plugins
            for enabled_plugin in self.plugins.keys():
                if enabled_plugin == __name__:
                    continue  # Skip ourself

                if not plugins or enabled_plugin not in plugins:
                    logger.info('Disabling plugin %s', enabled_plugin)
                    del self.plugins[enabled_plugin]

        if plugins:
            for plugin, cls in plugins.items():
                if init or plugin not in self.plugins:
                    logger.info('Initializing plugin %s', plugin)
                    self.plugins[plugin] = cls(config, reinit=False)
                else:
                    logger.info('Reloading plugin %s', plugin)
                    del self.plugins[plugin]
                    self.plugins[plugin] = cls(config, reinit=True)
                # xmpp attribute pointing to this instance is available in plugin object
                setattr(self.plugins[plugin], 'xmpp', self)

        cmds = self.available_commands(reset=True)
        logger.info('Registered commands: %s', ', '.join(cmds))

    def _room_members(self):
        """
        Change multi-user chat room member list.
        """
        query = ET.Element('{http://jabber.org/protocol/muc#admin}query')
        qitem = '{http://jabber.org/protocol/muc#admin}item'
        query.append(ET.Element(qitem, {'affiliation': 'owner', 'jid': self.boundjid.bare}))

        for jid in self.room_users:
            if jid in self.room_admins:
                affiliation = 'admin'
            else:
                affiliation = 'member'

            item = ET.Element(qitem, {'affiliation': affiliation, 'jid': jid})
            query.append(item)

        iq = self.make_iq_set(query)
        iq['to'] = self.room
        iq['from'] = ''
        iq.send()

    def _room_config(self):
        """
        Configure multi-user chat room.
        """
        logger.info('Getting current configuration for MUC room %s', self.room)

        try:
            self.room_config = self.muc.getRoomConfig(self.room)
        except ValueError:
            logger.error('Could not get MUC room configuration. Maybe the room is not (properly) initialized.')
            return

        if self.room_users:
            self.room_config['fields']['muc#roomconfig_membersonly']['value'] = True
            self.room_config['fields']['members_by_default']['value'] = False
        else:
            self.room_config['fields']['muc#roomconfig_membersonly']['value'] = False
            self.room_config['fields']['members_by_default']['value'] = True

        logger.info('Setting new configuration for MUC room %s', self.room)
        self.muc.setRoomConfig(self.room, self.room_config)
        logger.info('Setting member list for MUC room %s', self.room)
        self._room_members()

    def _jid_in_room(self, jid):
        """
        Determine if jid is present in chat room.
        """
        for nick in self.muc.rooms[self.room]:
            entry = self.muc.rooms[self.room][nick]

            if entry is not None and entry['jid'].bare == jid:
                return True

        return False

    def _handle_new_subscription(self, pres):
        """
        xmpp.auto_authorize is True by default, which is fine. But we want to
        restrict this to users only (if set). We do this by overriding the
        automatic subscription mechanism.
        """
        user = pres['from']

        if not self.users or user in self.users:
            logger.info('Allowing user "%s" to auto subscribe', user)
            return super(LudolphBot, self)._handle_new_subscription(pres)
        else:
            logger.warning('User "%s" is not allowed to subscribe', user)

    def get_jid(self, msg, bare=True):
        """
        Helper method for retrieving jid from message.
        """
        if msg['type'] == 'groupchat' and self.room:
            jid = self.muc.getJidProperty(msg['mucroom'], msg['mucnick'], 'jid')
        else:
            jid = msg['from']

        if bare and jid:
            return jid.bare

        return jid

    def session_start(self, event):
        """
        Process the session_start event.
        """
        self.get_roster()
        self.roster_cleanup()
        self.send_presence(pnick=self.nick)

        if self.room and self.muc:
            logger.info('Initializing multi-user chat room %s', self.room)
            self.muc.joinMUC(self.room, self.nick, maxhistory=self.maxhistory)

    def roster_cleanup(self):
        """
        Remove roster items with none subscription.
        """
        roster = self.client_roster
        logger.info('Current roster: %s', ', '.join(roster.keys()))

        # Remove users with none subscription from roster
        # Also remove users that are not in users setting (if set)
        for i in roster.keys():
            if roster[i]['subscription'] == 'none' or (self.users and i not in self.users):
                logger.warning('Roster item: %s (%s) - removing!', i, roster[i]['subscription'])
                self.send_presence(pto=i, ptype='unsubscribe')
                self.del_roster_item(i)
            else:
                logger.info('Roster item: %s (%s) - ok', i, roster[i]['subscription'])

    def muc_online(self, presence):
        """
        Process a presence stanza from a chat room.
        """
        # Configure room and say hello from jabber bot if this is a presence stanza
        if presence['from'] == '%s/%s' % (self.room, self.nick):
            self._room_config()
            self.msg_send(self.room, '%s is here!' % self.nick, mtype='groupchat')
            self._muc_ready = True
            self.send_presence(pto=presence['from'])
            logger.info('People in MUC room: %s', ', '.join(self.muc.getRoster(self.room)))

            # Send invitation to all users
            for user in self.room_users:
                if self._jid_in_room(user):
                    logger.info('User "%s" already in MUC room', user)
                elif user != self.room:
                    logger.info('Inviting "%s" to MUC room', user)
                    self.muc.invite(self.room, user)

        else:
            # Say hello to new user
            muc = presence['muc']
            logger.info('User "%s" with nick "%s", role "%s" and affiliation "%s" is joining MUC room',
                        muc['jid'], muc['nick'], muc['role'], muc['affiliation'])
            self.msg_send(presence['from'].bare, 'Hello %s!' % muc['nick'], mtype='groupchat')

    def available_commands(self, reset=False):
        """
        List of all available bot commands.
        """
        # Remove commands from disabled plugins
        if reset:
            for cmd_name, cmd in self.commands.items():
                if cmd['module'] not in self.plugins:
                    del self.commands[cmd_name]
            self._commands = None

        # Sort and cache
        if self._commands is None:
            self._commands = sorted(self.commands.keys())

        return self._commands

    def get_command(self, cmdstr):
        """
        Find text in available commands and return command dict.
        """
        if not cmdstr:
            return None

        if cmdstr in self.available_commands():
            cmd = self.commands[cmdstr]
        else:
            for key in self.available_commands():
                if key.startswith(cmdstr):
                    cmd = self.commands[key]
                    cmdstr = key
                    break
            else:
                return None

        cmd['str'] = cmdstr

        return cmd

    def message(self, msg, types=('chat', 'normal')):
        """
        Incoming message handler.
        """
        if msg['type'] not in types:
            return

        # Seek received text in available commands and get command
        cmd = self.get_command(msg['body'].split()[0].strip())

        if cmd:
            start_time = time.time()
            f_cmd = getattr(self.plugins[cmd['module']], cmd['name'])
            # Run command
            out = f_cmd(msg)

            if out:
                cmd_time = time.time() - start_time
                logger.info('Command %s.%s finished in %g seconds', cmd['module'], cmd['name'], cmd_time)

            return out
        else:
            # Send message that command was not understood and what to do
            return msg.reply('Sorry, I don\'t understand "%s"\n'
                             'Please type "help" for more info' % msg['body']).send()

    def muc_message(self, msg):
        """
        MUC Incoming message handler.
        """
        if not self._muc_ready:
            return

        if msg['mucnick'] == self.nick:
            return

        # Respond to the message only if the bots nickname is mentioned
        # And only if we can get user's JID
        nick = self.nick + ':'
        if msg['body'].startswith(nick) and self.get_jid(msg):
            msg['body'] = msg['body'].lstrip(nick).lstrip()
            return self.message(msg, types=('groupchat',))

    def shutdown(self, signalnum, handler):
        """
        Shutdown signal handler.
        """
        logger.info('Requested shutdown (%s)', signalnum)

        return self.abort()

    def reload(self, config, plugins=None):
        """
        Reload bot configuration and plugins.
        """
        logger.info('Requested reload')
        self._load_config(config)
        self._load_plugins(config, plugins, init=False)

        if self.room and self.muc:
            self._muc_ready = False
            self.muc.leaveMUC(self.room, self.nick)
            logger.info('Reinitializing multi-user chat room %s', self.room)
            self.muc.joinMUC(self.room, self.nick, maxhistory=self.maxhistory)

    def mon_thread(self):
        """
        Processing input from the monitoring pipe file.
        """
        with os.fdopen(os.open(self.pipe_file, os.O_RDONLY | os.O_NONBLOCK)) as fifo:
            logger.info('Processing input from monitoring pipe file')
            while not self.stop.is_set():
                line = fifo.readline().strip()
                if line:
                    data = line.split(';', 1)
                    if len(data) == 2:
                        logger.info('Sending monitoring message to "%s"', data[0])
                        logger.debug('\twith body: "%s"', data[1])

                        if data[0] == self.room:
                            mtype = 'groupchat'
                        else:
                            mtype = 'normal'

                        self.msg_send(data[0], data[1], mtype=mtype)

                    else:
                        logger.warning('Bad message format ("%s")', line)

                time.sleep(1)

                if self.stop.is_set():
                    self._end_thread('mon_thread', early=True)
                    return

    def msg_send(self, mto, mbody, **kwargs):
        """
        Create message and send it.
        """
        return LudolphMessage.create(mbody, **kwargs).send(self, mto)

    def msg_reply(self, msg, mbody, **kwargs):
        """
        Set message reply text and html, and send it.
        """
        if mbody is None:
            return None  # Command performs custom message sending

        return LudolphMessage.create(mbody, **kwargs).reply(msg)



"""
Ludolph: Monitoring Jabber bot
Copyright (C) 2012-2013 Erigones s. r. o.
This file is part of Ludolph.

See the file LICENSE for copying permission.
"""
import json
import logging
import requests

from ludolph.command import command, parameter_required, admin_required
from ludolph.message import red, green, blue
from ludolph.plugins.plugin import LudolphPlugin

logger = logging.getLogger(__name__)


class Erigones(LudolphPlugin):
    """
    Erigones API commands. EXPERIMENTAL.

    https://my.erigones.com/static/api/doc/

    Add to ludolph.cfg::

        [erigones]
        api_url = https://my.erigones.com/api
        username = username@example.com
        password = Passw0rd

    """

    user_agent = 'es/0.3/ludolph'
    methods = {
        'get':      requests.get,
        'create':   requests.post,
        'set':      requests.put,
        'delete':   requests.delete,
        'options':  requests.options,
    }
    actions = methods.keys()
    timeout = None
    headers = {
        'User-Agent': user_agent,
        'Accept': 'application/json; indent=4',
        'Content-Type': 'application/json; indent=4',
    }

    def __init__(self, config, **kwargs):
        """
        Initialize configuration and login to erigones.
        """
        self.api_url = config.get('erigones', 'api_url').rstrip('/')
        self.credentials = {
            'username': config.get('erigones', 'username'),
            'password': config.get('erigones', 'password'),
        }
        self._login()

    def __es(self, action, resource, params=None, data=None, msg=None):
        """
        The es http request. Returns (status_code, response_text).
        """
        url = self.api_url + resource
        response_text = ''
        r = self.methods[action](url=url, headers=self.headers, timeout=self.timeout,
                                 params=params, data=data, allow_redirects=False, stream=True)
        status_code = r.status_code

        if 'task_id' in r.headers:  # Task status stream
            msg_body = 'Waiting for pending task %s ...' % blue(r.headers['task_id'])
            self.xmpp.msg_send(msg['from'].bare, msg_body, mtype=msg['type'])

            for i in r.iter_content():
                if not i.isspace():
                    text = i + r.text
                    text = text.strip().split('\n')
                    status_code = int(text.pop())
                    response_text = '\n'.join(text)
                    break

        else:
            response_text = r.text

        try:
            response_text = json.loads(response_text)
        except ValueError:
            pass

        return status_code, response_text

    def _logout(self):
        """
        Logout from Erigones API.
        """
        logger.debug('Signing out of Erigones API')
        status, text = self.__es('get', '/accounts/logout/')

        if status == 200:
            logger.info('Logout successful: "%s"', text)
            self.headers.pop('Authorization', None)
        else:
            logger.warning('Logout problem (%s): "%s"', status, text)

    def _login(self):
        """
        Login to Erigones API.
        """
        logger.debug('Signing in to Erigones API')
        self.headers.pop('Authorization', None)
        status, text = self.__es('create', '/accounts/login/', data=json.dumps(self.credentials))

        if status == 200 and isinstance(text, dict) and 'token' in text:
            logger.info('Login successful: "%s"', text['detail'])
            self.headers['Authorization'] = 'Token ' + text['token']
        else:
            logger.error('Login problem (%s): "%s"', status, text)

    def _is_authenticated(self):
        """
        Return True if authorization token exists.
        """
        return 'Authorization' in self.headers

    def _es(self, msg, action, resource, *parameters):
        """
        The es command. Returns (status_code, response_text).
        """
        if action not in self.actions:
            return 0, 'ERROR: Bad action'

        if not resource or resource[0] != '/':
            return 0, 'ERROR: Missing resource'

        params = None
        data = None
        options = {}

        if parameters:  # Parse arguments
            key = None
            val_next = False

            for i in parameters:
                if i and i[0] == '-':
                    key = i[1:]
                    options[key] = True
                    val_next = True
                    continue

                if val_next and key:
                    _i = str(i).lower()

                    if _i == 'false':
                        options[key] = False
                    elif _i == 'true':
                        options[key] = True
                    elif _i == 'null':
                        options[key] = None
                    else:
                        options[key] = i

                    key = None
                    val_next = False

        if action == 'get' or action == 'logout':
            params = options
        else:
            data = json.dumps(options)

        if resource[-1] != '/':
            resource += '/'

        def workbitch():  # Perform one re-login if needed
            if not self._is_authenticated():
                logger.error('Erigones API not available')
                return 0, 'ERROR: Erigones API not available'

            return self.__es(action, resource, params=params, data=data, msg=msg)

        status, res = workbitch()

        if status == 403 and res.get('detail') == 'Authentication credentials were not provided.':
            logger.warning('Performing re-login to Erigones API')
            self._login()
            return workbitch()

        return status, res

    @admin_required
    @parameter_required(2)
    @command
    def es(self, msg, action, resource, *parameters):
        """
        es - Swiss Army Knife for Erigones API (EXPERIMENTAL)

        Usage: es action [/resource] [parameters]

          action:\t{get|create|set|delete|options}
          resource:\t/some/resource/in/api
          parameters:\t-foo baz -bar qux ...
        """
        status, text = self._es(msg, action, resource, *parameters)

        if not status:
            return text

        out = {
            'action': action,
            'resource': resource,
            '**status**': status,
            '**text**': text,
        }

        return json.dumps(out, indent=4)

    @admin_required
    @command
    def vm_list(self, msg):
        """
        Show a list of all servers.

        Usage: vm-list
        """
        code, res = self._es(msg, 'get', '/vm/status')

        if code != 200:
            return str(res)

        out = ''

        for vm in res['result']:
            if vm['status'] == 'running':
                color = green
            elif vm['status'] == 'stopped' or vm['status'] == 'stopping':
                color = red
            elif vm['status'] == 'pending':
                color = blue
            else:
                color = lambda x: x

            out += '**%s** (%s)\t%s\t(%d)\n' % (vm['hostname'], vm['alias'], color(vm['status']), len(vm['tasks']))

        out += '\n**%d** servers are shown.' % len(res['result'])

        return out



"""
Ludolph: Monitoring Jabber bot
Copyright (C) 2012-2013 Erigones s. r. o.
This file is part of Ludolph.

See the file LICENSE for copying permission.
"""
import logging
from re import compile as r
from sleekxmpp.xmlstream import ET
try:
    from xml.etree.ElementTree import ParseError
except ImportError:
    from xml.parsers.expat import ExpatError as ParseError

logger = logging.getLogger(__name__)

__all__ = ['red', 'green', 'blue', 'LudolphMessage']

TEXT2BODY = [
        (r(r'\*\*(.+?)\*\*'), r'*\1*'),
        (r(r'__(.+?)__'), r'\1'),
        (r(r'\^\^(.+?)\^\^'), r'\1'),
        (r(r'~~(.+?)~~'), r'\1'),
        (r(r'%{(.+?)}(.+)%'), r'\2'),
]

TEXT2HTML = [
        ('&', '&#38;'),
        ('<', '&#60;'),
        ('>', '&#62;'),
        ("'", '&#39;'),
        ('"', '&#34;'),
        (r(r'\*\*(.+?)\*\*'), r'<b>\1</b>'),
        (r(r'__(.+?)__'), r'<i>\1</i>'),
        (r(r'\^\^(.+?)\^\^'), r'<sup>\1</sup>'),
        (r(r'~~(.+?)~~'), r'<sub>\1</sub>'),
        (r(r'%{(.+?)}(.+)%'), r'<span style="\1">\2</span>'),
        (r(r'(ERROR)'), r'<span style="color:#FF0000;">\1</span>'),
        (r(r'(PROBLEM|OFF)'), r'<span style="color:#FF0000;"><strong>\1</strong></span>'),
        (r(r'(OK|ON)'), r'<span style="color:#00FF00;"><strong>\1</strong></span>'),
        (r(r'([Dd]isaster)'), r'<span style="color:#FF0000;"><strong>\1</strong></span>'),
        (r(r'([Cc]ritical)'), r'<span style="color:#FF3300;"><strong>\1</strong></span>'),
        (r(r'([Hh]igh)'), r'<span style="color:#FF6600;"><strong>\1</strong></span>'),
        (r(r'([Aa]verage)'), r'<span style="color:#FF9900;"><strong>\1</strong></span>'),
        (r(r'([Ww]arning)'), r'<span style="color:#FFCC00;"><strong>\1</strong></span>'),
        #(r(r'([Ii]nformation)'), r'<span style="color:#FFFF00;"><strong>\1</strong></span>'),
        (r(r'(Monitored)'), r'<span style="color:#00FF00;"><strong>\1</strong></span>'),
        (r(r'(Not\ monitored)'), r'<span style="color:#FF0000;"><strong>\1</strong></span>'),
        ('\n', '<br/>\n'),
]


def red(s):
    return '%%{color:#FF0000}%s%%' % s


def green(s):
    return '%%{color:#00FF00}%s%%' % s


def blue(s):
    return '%%{color:#0000FF}%s%%' % s


class LudolphMessage(object):
    """
    Creating and sending bot's messages (replies).
    """
    mbody = None
    mhtml = None
    mtype = None

    def __init__(self, mbody, mhtml=None, mtype=None):
        """
        Construct message body in plain text and html.
        """
        self.mtype = mtype

        if mbody is not None:
            self.mbody = self._text2body(str(mbody))

        if mhtml is None and mbody is not None:
            self.mhtml = self._text2html(str(mbody))
        else:
            self.mhtml = str(mhtml)

    def _replace(self, replist, text):
        """
        Helper for replacing text parts according to replist.
        """
        for r, t in replist:
            if isinstance(r, basestring):
                text = text.replace(r, t)
            else:
                text = r.sub(t, text)

        return text

    def _text2body(self, text):
        """
        Remove tags from text.
        """
        body = self._replace(TEXT2BODY, text.strip())

        return body

    def _text2html(self, text):
        """
        Convert text to html.
        """
        html = self._replace(TEXT2HTML, text.strip())
        html = '<div>\n' + html + '\n</div>'

        try:
            return ET.XML(html)
        except (ParseError, SyntaxError) as e:
            logger.error('Could not parse html: %s', e)
            return None

    @classmethod
    def create(cls, mbody, **kwargs):
        """
        Return LudolphMessage instance.
        """
        if isinstance(mbody, cls):
            return mbody

        return cls(mbody, **kwargs)

    def send(self, xmpp, mto):
        """
        Send a new message.
        """
        msg = xmpp.make_message(mto, self.mbody, mtype=self.mtype, mhtml=self.mhtml)

        return msg.send()

    def reply(self, msg, clear=True):
        """
        Send a reply to incoming msg.
        """
        msg.reply(self.mbody, clear=clear)
        msg['html']['body'] = self.mhtml

        return msg.send()
